using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Threading.Tasks;
using MaynotPersistence;

namespace MaynotModel
{
    public class MaynotGameModel
    {
		#region Fields
		private IPersistence _dataAccess;
        private MaynotGameState _state;
        private int debug;
        #endregion

        #region Events
        /// <summary>
        /// Event of saving game.
        /// </summary>
        public event EventHandler<MaynotEventArg>? SaveGame;
        /// <summary>
        /// Event of loading game.
        /// </summary>
        public event EventHandler<MaynotEventArg>? LoadGame;
        /// <summary>
        /// Event of exiting game.
        /// </summary>
        public event EventHandler<MaynotEventArg>? ExitGame;
        /// <summary>
        /// Event of placing a <see cref="Tile"/>.
        /// </summary>
        public event EventHandler<MaynotEventArg>? TilePlaced;
        /// <summary>
        /// Event of information asked.
        /// </summary>
        public event EventHandler<MaynotEventArg>? GeneralInfoAsked;
        /// <summary>
        /// Event of new day.
        /// </summary>
        public event EventHandler<TimeElapsedEventArgs>? DayElapsed;
        /// <summary>
        /// Event of inspecting a new <see cref="Tile"/>.
        /// </summary>
        public event EventHandler<InspectedTileChangedEventArgs>? InspectedTileChanged;
        /// <summary>
        /// Event of catatrophe.
        /// </summary>
        public event EventHandler<MaynotEventArg>? catastropheHappened;
        /// <summary>
        /// Event of new game speed.
        /// </summary>
        public event EventHandler<EventArgs>? GameSpeedChanged;
        /// <summary>
        /// Event of changed citizens.
        /// </summary>
        public event EventHandler<EventArgs>? UpdatePopulation;
        /// <summary>
        /// Event of loading game finished.
        /// </summary>
        public event EventHandler<MaynotEventArg>? GameLoaded;
        /// <summary>
        /// Event of losing a game.
        /// </summary>
        public event EventHandler<EventArgs>? GameOver;
        /// <summary>
        /// Event of new <see cref="Transaction"/>.
        /// </summary>
        public event EventHandler<TransactionEventArgs>? TransactionHappenned;
        #endregion

        #region Properties
        /// <summary>
        /// Budget of city.
        /// </summary>
        public float Money { get => _state.money; set => _state.money = value; }
        /// <summary>
        /// Speed of game.
        /// </summary>
        public float Speed { get { return _state.gameSpeed; }  }
        /// <summary>
        /// Number of citizens.
        /// </summary>
        public int Population { get { return _state.citizens.Count; } }
        /// <summary>
        /// Residential tax precantage.
        /// </summary>
        public int ResidentalTax { get { return _state.ResidentalTax; } }
        /// <summary>
        /// Service tax precantage.
        /// </summary>
        public int ServiceTax { get { return _state.ServiceTax; } }
        /// <summary>
        /// Industrial tax precantage.
        /// </summary>
        public int IndustrialTax { get { return _state.IndustrialTax; } }
        /// <summary>
        /// Currently selected <see cref="Tile"/> to inspect.
        /// </summary>
        public Tile InspectedTile { get; set; }
        /// <summary>
        /// Average satisfaction of citizens.
        /// </summary>
        public double AverageSatisfaction { get { return _state.averageSatisfaction; } }
        /// <summary>
        /// <see cref="DateTime"/> of game.
        /// </summary>
        public DateTime Time { get => _state.time; set => _state.time = value; }
        /// <summary>
        /// Speed of game.
        /// </summary>
        public int GameSpeed { get => _state.gameSpeed; set => _state.gameSpeed = value; }
        /// <summary>
        /// Game table board.
        /// </summary>
        public Tile[,] GameBoard { get => _state.gameBoard; set => _state.gameBoard = value; }
        /// <summary>
        /// Size of <see cref="GameBoard"/>.
        /// </summary>
        public int GameBoardSize { get => _state.size; }
        /// <summary>
        /// List of citizens.
        /// </summary>
        public List<Person> Citizens { get => _state.citizens; set => _state.citizens = value; }
        /// <summary>
        /// List of saved transactions.
        /// </summary>
        public List<Transaction> Transactions { get; set; }
		public int GetStartRoadX()
		{
			return _state.startX;
		}
		public int GetStartRoadY()
		{
			return _state.startY;
		}
        #endregion

        #region Constructors
        /// <summary>
        /// Construct new <see cref="MaynotModel"/>.
        /// </summary>
        /// <param name="dataAccess">persistence object</param>
        public MaynotGameModel(IPersistence dataAccess)
        {
            _dataAccess = dataAccess;
			_state = new MaynotGameState();
            debug = 0;
		}
        #endregion

        #region New Game
        /// <summary>
        /// Sets model for new game.
        /// </summary>
        /// <remarks>
        /// Creates new <see cref="MaynotGameState"/>, <see cref="Transactions"/> and calls <see cref="ForestGenerator"/> and <see cref="OnGameLoaded"/>.
        /// </remarks>
        public void NewGame()
		{
			InspectedTile = null;
			_state.timer.Elapsed -= Timer_Elapsed;
			_state = new MaynotGameState(30);
			_state.timer.Elapsed += Timer_Elapsed;
            Transactions = new List<Transaction>();
			ForestGenerator();
			OnGameLoaded();
			Debug.WriteLine("New game");
		}
		private void ForestGenerator()
		{
			Random _r = new();
			int counter = _r.Next(30);
			double distanceFromStartCenter;

			for (int i = 0; i < counter; i++)
			{
				int x;
				int y;
				do
				{
					x = _r.Next(GameBoardSize);
					y = _r.Next(GameBoardSize);
					distanceFromStartCenter = Math.Sqrt((x - _state.startX) * (x - _state.startX) + (y - _state.startY) * (y - _state.startY));
				} while (GameBoard[x, y] is not Empty && distanceFromStartCenter > 8);

				int area = _r.Next(1, 10);
				for (int k = x - area; k <= x + area; k++)
				{
					for (int l = y - area; l <= y + area; l++)
					{
						double distanceFromStart = Math.Sqrt((k - _state.startX) * (k - _state.startX) + (l - _state.startY) * (l - _state.startY));
						if (_state.IsValid(k, l) && GameBoard[k, l] is Empty && distanceFromStart > 8)
						{
							int ap = _r.Next(area);
							double distance = Math.Sqrt((x - k) * (x - k) + (y - l) * (y - l));
							if (distance < area / 2)
							{
								GameBoard[k, l] = new Forest(Time, true);
							}
							else if (distance < ap)
							{
								GameBoard[k, l] = new Forest(Time, true);
							}
						}
					}
				}
			}
		}
		#endregion        

		#region Time Managments
		private void Timer_Elapsed(object? sender, System.Timers.ElapsedEventArgs e)
		{
			_state.time = _state.time.AddDays(1);  
            growForests();
            buildingEffects();
            CalculateAverageSatisfaction();
            
			if ((int)_state.time.DayOfWeek % 3 == 0) //Minden harmadik nap futó tevékenységek
			{
				MovingOut();
				OnInspectedTileChanged(); //Csak ilyenkor frissítjük az inspected-et, hogy kevesebbet flicker-ezzen
			}

			if ((int)_state.time.DayOfWeek == 1)
			{
				MovingIn();
			}

			if (_state.time.Day == 1) 
			{
				catastrophe();
                if (_state.time.Year !=1 && _state.time.Month != 1)
                {
                    lose();
                }
			}

			if (_state.time.Month == 1 && _state.time.Day == 1)
			{
                PayExpenses();
				taxCollection();
				managingPension();
				aging();
				SetEducationLevel();
				EndLifes();
				negativeMoney();
                highTaxes();
                lowTaxes();
            }
            OnTimeElapsed();                        
        }
        /// <summary>
        /// Stops <see cref="System.Timers.Timer"/> in <see cref="_state"/>
        /// </summary>
        /// <remarks>
        /// Invokes <see cref="GameSpeedChanged"/>.
        /// </remarks>
        public void stopTime()
        {
            if (_state.gameSpeed == 0) return;
            _state.timer.Stop();
            _state.prevGameSpeed = _state.gameSpeed;
            _state.gameSpeed = 0;
            GameSpeedChanged?.Invoke(this, new EventArgs());
        }
        /// <summary>
        /// Resumes <see cref="System.Timers.Timer"/> in <see cref="_state"/>
        /// </summary>
        /// <remarks>
        /// Invokes <see cref="GameSpeedChanged"/>.
        /// </remarks>
        public void resumeTime()
        {
			if (_state.gameSpeed != 0) return;
			if (_state.gameSpeed == 0)
            {
				_state.timer.Start();
                _state.gameSpeed = 1;
				GameSpeedChanged?.Invoke(this, new EventArgs());
			}
            else
            {
				_state.timer.Start();
				_state.gameSpeed = _state.prevGameSpeed;
				GameSpeedChanged?.Invoke(this, new EventArgs());
			}
        }
        /// <summary>
        /// Slows down <see cref="System.Timers.Timer"/> in <see cref="_state"/>
        /// </summary>
        public void slowTime()
        {
            switch (_state.gameSpeed)
            {
                case 3:
                    _state.gameSpeed = 2;
                    _state.timer.Stop();
                    _state.timer.Interval = 500 / _state.gameSpeed;
                    _state.timer.Start();
                    break;
                case 2:
                    _state.gameSpeed = 1;
                    _state.timer.Stop();
                    _state.timer.Interval = 500 / _state.gameSpeed;
                    _state.timer.Start();
                    break;
                case 1:
                    _state.gameSpeed = 0;
                    _state.timer.Stop();                   
                    break;
            }
            GameSpeedChanged?.Invoke(this, new EventArgs());
            Debug.WriteLine("Gamespeed: " + GameSpeed);
        }
        /// <summary>
        /// Speeds up <see cref="System.Timers.Timer"/> in <see cref="_state"/>
        /// </summary>
        public void speedUpTime()
        {
            switch (_state.gameSpeed)
            {
                case 0:
                    _state.gameSpeed = 1;
                    _state.timer.Stop();
                    _state.timer.Interval = 500 / _state.gameSpeed;
                    _state.timer.Start();
                    break;
                case 1:
                    _state.gameSpeed = 2;
                    _state.timer.Stop();
                    _state.timer.Interval = 500 / _state.gameSpeed;
                    _state.timer.Start();
                    break;
                case 2:
                    _state.gameSpeed = 3;
                    _state.timer.Stop();
                    _state.timer.Interval = 500 / _state.gameSpeed;
                    _state.timer.Start();
                    break;
            }
            GameSpeedChanged?.Invoke(this, new EventArgs());
            Debug.WriteLine("Gamespeed: " + GameSpeed);
        }
		#endregion

		#region Place Tile
		private bool placeTile(Tile t, int x, int y)
        {
            if (_state.gameBoard[x, y] is Empty)
            {
                _state.gameBoard[x, y] = t;
                if (t is ResidentialZone)
                {
                    _state.residentalZones.Add((x, y));
                }
                else if (t is IndustrialZone)
                {
                    _state.industrialZones.Add((x, y));
                }
                else if (t is ServiceZone)
                {
                    _state.serviceZones.Add((x, y));
                }
                return true;
            }
            else
            {
                return false;
            }


        }
        /// <summary>
        /// Creates a <see cref="Road"/> by location.
        /// </summary>
        /// <remarks>
        /// Places the <see cref="Road"/>, calls <see cref="UpdateMaintenanceFees"/> and updates <see cref="Transactions"/> accordingly.
        /// </remarks>
        /// <param name="x">row number of <see cref="Tile"/> in <see cref="GameBoard"/></param>
        /// <param name="y">column number of <see cref="Tile"/> in <see cref="GameBoard"/></param>
        /// <returns>the succession of placing a <see cref="Road"/> in given position.</returns>
        public bool placeRoad(int x, int y)
        {
            if (_state.gameBoard[x, y] is Empty)
            {
                Road r = new Road();
                _state.money = _state.money - Road.buildCost;
                UpdateTransactions(new Transaction(TransactionType.BUILDING, Road.buildCost, Time, r.GetType()));
                UpdateMaintenanceFees(r, Road.maintenanceFee);
                return placeTile(r, x, y);
            }
            else { return false; }
            
        }
        /// <summary>
        /// Creates a <see cref="Forest"/> by location.
        /// </summary>
        /// <remarks>
        /// Places the <see cref="Forest"/>, calls <see cref="UpdateMaintenanceFees(Tile, float)"/> and updates <see cref="Transactions"/> accordingly.
        /// </remarks>
        /// <param name="x">row number of <see cref="Tile"/> in <see cref="GameBoard"/></param>
        /// <param name="y">column number of <see cref="Tile"/> in <see cref="GameBoard"/></param>
        /// <returns>the succession of placing a <see cref="Forest"/> in given position.</returns>
        public bool placeForest(int x, int y)
        {
            if (_state.gameBoard[x, y] is Empty)
            {
                Forest f = new Forest(_state.time, false);
                _state.money -= Forest.BuildCost;
                UpdateTransactions(new Transaction(TransactionType.BUILDING, Forest.BuildCost, Time, f.GetType()));
                UpdateMaintenanceFees(f, Forest.MaintenanceFee);
                return placeTile(f, x, y);
            }
            else { return false; }          
        }
        /// <summary>
        /// Creates a <see cref="Stadium"/> by location.
        /// </summary>
        /// <remarks>
        /// Places the <see cref="Stadium"/>, calls <see cref="UpdateMaintenanceFees(Tile, float)"/> and updates <see cref="Transactions"/> accordingly.
        /// </remarks>
        /// <param name="x">row number of <see cref="Tile"/> in <see cref="GameBoard"/></param>
        /// <param name="y">column number of <see cref="Tile"/> in <see cref="GameBoard"/></param>
        /// <returns>the succession of placing a <see cref="Stadium"/> in given position.</returns>
        public bool placeStadium(int x, int y)
        {
            Random rand = new Random();
            int id = rand.Next(1, 1000);
            Stadium s = new Stadium(id);
            _state.money = _state.money - Stadium.buildCost;
            UpdateTransactions(new Transaction(TransactionType.BUILDING, Stadium.buildCost, Time, s.GetType()));
            UpdateMaintenanceFees(s, Stadium.maintenanceFee);
            if (_state.gameBoard[x, y] is Empty && x + 1 < _state.size && y + 1 < _state.size && _state.gameBoard[x + 1, y] is Empty && _state.gameBoard[x, y + 1] is Empty && _state.gameBoard[x + 1, y + 1] is Empty)
            {
                _state.gameBoard[x, y] = s;
                _state.gameBoard[x+1, y] = s;
                _state.gameBoard[x, y+1] = s;
                _state.gameBoard[x+1, y+1] = s;
                return true;
            }
            else
            {
                return false;
            }
        }
        /// <summary>
        /// Creates a <see cref="PoliceStation"/> by location.
        /// </summary>
        /// <remarks>
        /// Places the <see cref="PoliceStation"/>, calls <see cref="UpdateMaintenanceFees(Tile, float)"/> and updates <see cref="Transactions"/> accordingly.
        /// </remarks>
        /// <param name="x">row number of <see cref="Tile"/> in <see cref="GameBoard"/></param>
        /// <param name="y">column number of <see cref="Tile"/> in <see cref="GameBoard"/></param>
        /// <returns>the succession of placing a <see cref="PoliceStation"/> in given position.</returns>
        public bool placePoliceStation(int x, int y)
        {
            if (_state.gameBoard[x, y] is Empty)
            {
                PoliceStation p = new PoliceStation();
                _state.money = _state.money - PoliceStation.buildCost;
                UpdateTransactions(new Transaction(TransactionType.BUILDING, PoliceStation.buildCost, Time, p.GetType()));
                UpdateMaintenanceFees(p, PoliceStation.maintenanceFee);
                return placeTile(p, x, y);
            }
            else 
            { 
                return false; 
            }
        }
        /// <summary>
        /// Creates a <see cref="School"/> by location.
        /// </summary>
        /// <remarks>
        /// Places the <see cref="School"/>, calls <see cref="UpdateMaintenanceFees(Tile, float)"/> and updates <see cref="Transactions"/> accordingly.
        /// </remarks>
        /// <param name="x">row number of <see cref="Tile"/> in <see cref="GameBoard"/></param>
        /// <param name="y">column number of <see cref="Tile"/> in <see cref="GameBoard"/></param>
        /// <returns>the succession of placing a <see cref="School"/> in given position.</returns>
        public bool placeSchool(int x, int y)
        {
            Random rand = new Random();
            int id = rand.Next(1, 1000);
            School s = new School(id);
            _state.money = _state.money - School.buildCost;
            UpdateTransactions(new Transaction(TransactionType.BUILDING, School.buildCost, Time, s.GetType()));
            UpdateMaintenanceFees(s, School.maintenanceFee);
            if (_state.gameBoard[x, y] is Empty && x + 1 < _state.size && y + 1 < _state.size && _state.gameBoard[x + 1, y] is Empty)
            {
                _state.gameBoard[x, y] = s;
                _state.gameBoard[x + 1, y] = s;
                return true;
            }
            else
            {
                return false;
            }
        }
        /// <summary>
        /// Creates a <see cref="University"/> by location.
        /// </summary>
        /// <remarks>
        /// Places the <see cref="University"/>, calls <see cref="UpdateMaintenanceFees(Tile, float)"/> and updates <see cref="Transactions"/> accordingly.
        /// </remarks>
        /// <param name="x">row number of <see cref="Tile"/> in <see cref="GameBoard"/></param>
        /// <param name="y">column number of <see cref="Tile"/> in <see cref="GameBoard"/></param>
        /// <returns>the succession of placing a <see cref="University"/> in given position.</returns>
        public bool placeUni(int x, int y)
        {
            Random rand = new Random();
            int id = rand.Next(1, 1000);
            University u = new University(id);
            _state.money = _state.money - University.buildCost;
            UpdateTransactions(new Transaction(TransactionType.BUILDING, University.buildCost, Time, u.GetType()));
            UpdateMaintenanceFees(u, University.maintenanceFee);
            if (_state.gameBoard[x, y] is Empty &&  x + 1 < _state.size && y + 1 < _state.size && _state.gameBoard[x + 1, y] is Empty && _state.gameBoard[x , y + 1] is Empty && _state.gameBoard[x + 1, y + 1] is Empty)
            {
                _state.gameBoard[x, y] = u;
                _state.gameBoard[x + 1, y] = u;
                _state.gameBoard[x, y + 1] = u;
                _state.gameBoard[x + 1, y + 1] = u;
                return true;
            }
            else
            {
                return false;
            }
        }
        /// <summary>
        /// Creates a <see cref="ResidentialZone"/> by location.
        /// </summary>
        /// <remarks>
        /// Places the <see cref="ResidentialZone"/> with <see cref="placeTile"/>.
        /// </remarks>
        /// <param name="x">row number of <see cref="Tile"/> in <see cref="GameBoard"/></param>
        /// <param name="y">column number of <see cref="Tile"/> in <see cref="GameBoard"/></param>
        /// <returns>the succession of placing a <see cref="ResidentialZone"/> in given position.</returns>
        public bool placeResidentialZone(int x, int y)
        {
            Debug.WriteLine("Placing residental in Model!");
            ResidentialZone r = new ResidentialZone(x, y, 0, new Random().Next(20, 40));
			return placeTile(r, x, y);
        }
        /// <summary>
        /// Creates a <see cref="IndustrialZone"/> by location.
        /// </summary>
        /// <remarks>
        /// Places the <see cref="ServiceZone"/> with <see cref="placeTile"/>.
        /// </remarks>
        /// <param name="x">row number of <see cref="Tile"/> in <see cref="GameBoard"/></param>
        /// <param name="y">column number of <see cref="Tile"/> in <see cref="GameBoard"/></param>
        /// <returns>the succession of placing a <see cref="ServiceZone"/> in given position.</returns>
        public bool placeIndustrialZone(int x, int y)
        {
            Debug.WriteLine("Placing industrial in Model");
            IndustrialZone i = new IndustrialZone(x, y, 0, new Random().Next(20, 40));
            return placeTile(i, x, y);
        }
        /// <summary>
        /// Creates a <see cref="ServiceZone"/> by location.
        /// </summary>
        /// <remarks>
        /// Places the <see cref="ServiceZone"/> with <see cref="placeTile"/>.
        /// </remarks>
        /// <param name="x">row number of <see cref="Tile"/> in <see cref="GameBoard"/></param>
        /// <param name="y">column number of <see cref="Tile"/> in <see cref="GameBoard"/></param>
        /// <returns>the succession of placing a <see cref="ServiceZone"/> in given position.</returns>
        public bool placeServiceZone(int x, int y)
        {
            Debug.WriteLine("Placing service in Model");
            ServiceZone s = new ServiceZone(x, y, 0, new Random().Next(20, 40));
			return placeTile(s, x, y);
        }
        #endregion

        #region Inspect Tile
        /// <summary>
        /// Selects a given <see cref="Tile"/> by location.
        /// </summary>
        /// <remarks>
        /// Sets <see cref="InspectedTile"/> and calls <see cref="OnInspectedTileChanged"/>.
        /// </remarks>
        /// <param name="x">row number of <see cref="Tile"/> in <see cref="GameBoard"/></param>
        /// <param name="y">column number of <see cref="Tile"/> in <see cref="GameBoard"/></param>
        public void InspectZone(int x, int y)
		{
			InspectedTile = _state.gameBoard[x, y];
			OnInspectedTileChanged();
		}
        /// <summary>
        /// Upgrades a given <see cref="Tile"/> by location if possible.
        /// </summary>
        /// <remarks>
        /// Sets <see cref="ZoneLevel"/> for <see cref="Zone"/>, calls <see cref="OnInspectedTileChanged"/> and updates <see cref="Transactions"/> accordingly.
        /// </remarks>
		public void UpgradeInspectedZone()
		{
			if (InspectedTile is Zone zone)
			{
                if(zone.Level != ZoneLevel.LARGE)
                {
				    zone.Upgrade();
				    _state.money -= zone.UpgradeCost;
                    UpdateTransactions(new Transaction(TransactionType.UPGRADE, zone.UpgradeCost, Time,zone.GetType()));
				    OnInspectedTileChanged();
                }
			}
		}
        #endregion

        #region Destroy Tile
        /// <summary>
        /// Destroys a given <see cref="Road"/> by location if possible.
        /// </summary>
        /// <param name="x">row number of <see cref="Tile"/> in <see cref="GameBoard"/></param>
        /// <param name="y">column number of <see cref="Tile"/> in <see cref="GameBoard"/></param>
        /// <returns>the succession of destroying the <see cref="Road"/> in given position.</returns>
        public bool DestroyRoad(int x, int y)
		{
			if (_state.gameBoard[x, y] is Road && _state.CanDestroyRoad(x, y))
			{
				_state.gameBoard[x, y] = new Empty();
				return true;
			}
			else
			{
				return false;
			}
		}
        /// <summary>
        /// Destroys a given <see cref="Tile"/> by location if possible.
        /// </summary>
        /// <remarks>
        /// Sets <see cref="GameBoard"/> item <see cref="Empty"/> if successful and updates <see cref="Transactions"/>, <see cref="_state.MaintenanceFees"/> accordingly.
        /// </remarks>
        /// <param name="x">row number of <see cref="Tile"/> in <see cref="GameBoard"/></param>
        /// <param name="y">column number of <see cref="Tile"/> in <see cref="GameBoard"/></param>
        /// <param name="byCatastrophe">flag for cause of calling</param>
        /// <returns>the succession of destroying the <see cref="Tile"/> in given position.</returns>
		public bool DestroyTile(int x, int y, bool byCatastrophe)
        {
            Tile tile = _state.gameBoard[x, y];
			if (tile is Road)
			{
				if (DestroyRoad(x, y)) // Le is törli, ha le lehet
				{
                    UpdateMaintenanceFees(tile, -Road.maintenanceFee);
                    _state.money += Road.buildCost / 2;
                    if (!byCatastrophe) Transactions.Add(new Transaction(TransactionType.BULLDOZER, Road.buildCost / 2, Time, tile.GetType()));
					return true;
				}
				else
				{
					return false;
				}
			}
			else if(tile is Zone zone)
            {
                if (zone.GetPeoples(_state.citizens).Count == 0)
                {
                    if (zone is ResidentialZone)
                        _state.residentalZones.Remove((x, y));

					if (zone is ServiceZone)
						_state.serviceZones.Remove((x, y));

					if (zone is IndustrialZone)
						_state.industrialZones.Remove((x, y));

					_state.gameBoard[x, y] = new Empty();
                    return true;
                }
                return false;
            }
            else if(tile is PoliceStation)
            {
                UpdateMaintenanceFees(tile, -PoliceStation.maintenanceFee);
                _state.money += PoliceStation.buildCost / 2;
                _state.gameBoard[x, y] = new Empty();
                if (!byCatastrophe) UpdateTransactions(new Transaction(TransactionType.BULLDOZER, PoliceStation.buildCost / 2, Time, tile.GetType()));
                return true;
            }
            else if (tile is Stadium)
            {
                UpdateMaintenanceFees(tile, -Stadium.maintenanceFee);
                _state.money += Stadium.buildCost / 2;
                if (!byCatastrophe) UpdateTransactions(new Transaction(TransactionType.BULLDOZER, Stadium.buildCost / 2, Time, tile.GetType()));
                Stadium tmpS = (Stadium)_state.gameBoard[x, y];
                int id = tmpS.Id;
                _state.gameBoard[x, y] = new Empty();

                if (_state.gameBoard[x + 1, y] is Stadium){
                    tmpS = (Stadium)_state.gameBoard[x + 1, y];
                    if (tmpS.Id == id) _state.gameBoard[x + 1, y] = new Empty();
                }

                if (_state.gameBoard[x - 1, y] is Stadium)
                {
                    tmpS = (Stadium)_state.gameBoard[x - 1, y];
                    if (tmpS.Id == id) _state.gameBoard[x - 1, y] = new Empty();
                }

                if (_state.gameBoard[x, y - 1] is Stadium )
                {
                    tmpS = (Stadium)_state.gameBoard[x, y - 1];
                    if (tmpS.Id == id) _state.gameBoard[x, y - 1] = new Empty();
                }

                if (_state.gameBoard[x, y + 1] is Stadium)
                {
                    tmpS = (Stadium)_state.gameBoard[x, y + 1];
                    if (tmpS.Id == id) _state.gameBoard[x, y + 1] = new Empty();
                }

                if (_state.gameBoard[x + 1, y + 1] is Stadium) 
                {
                    tmpS = (Stadium)_state.gameBoard[x + 1, y + 1];
                    if (tmpS.Id == id) _state.gameBoard[x + 1, y + 1] = new Empty();
                }

                if (_state.gameBoard[x + 1, y - 1] is Stadium)
                {
                    tmpS = (Stadium)_state.gameBoard[x + 1, y - 1];
                    if (tmpS.Id == id) _state.gameBoard[x + 1, y - 1] = new Empty();
                }

                if (_state.gameBoard[x - 1, y - 1] is Stadium)
                {
                    tmpS = (Stadium)_state.gameBoard[x - 1, y - 1];
                    if (tmpS.Id == id) _state.gameBoard[x - 1, y - 1] = new Empty();
                }

                if (_state.gameBoard[x - 1, y + 1] is Stadium)
                {
                    tmpS = (Stadium)_state.gameBoard[x - 1, y + 1];
                    if (tmpS.Id == id) _state.gameBoard[x - 1, y + 1] = new Empty();
                }               

                //A nagyobb cuccoknál ez nem ennyire egyszerű
                return true;
            }
            else if (tile is University)
            {
                UpdateMaintenanceFees(tile, -University.maintenanceFee);
                _state.money += University.buildCost / 2;
                if (!byCatastrophe) UpdateTransactions(new Transaction(TransactionType.BULLDOZER, University.buildCost / 2, Time, tile.GetType()));
                University tmpS = (University)_state.gameBoard[x, y];
                int id = tmpS.Id;
                _state.gameBoard[x, y] = new Empty();

                if (_state.gameBoard[x + 1, y] is University)
                {
                    tmpS = (University)_state.gameBoard[x + 1, y];
                    if (tmpS.Id == id) _state.gameBoard[x + 1, y] = new Empty();
                }

                if (_state.gameBoard[x - 1, y] is University)
                {
                    tmpS = (University)_state.gameBoard[x - 1, y];
                    if (tmpS.Id == id) _state.gameBoard[x - 1, y] = new Empty();
                }

                if (_state.gameBoard[x, y - 1] is University)
                {
                    tmpS = (University)_state.gameBoard[x, y - 1];
                    if (tmpS.Id == id) _state.gameBoard[x, y - 1] = new Empty();
                }

                if (_state.gameBoard[x, y + 1] is University)
                {
                    tmpS = (University)_state.gameBoard[x, y + 1];
                    if (tmpS.Id == id) _state.gameBoard[x, y + 1] = new Empty();
                }

                if (_state.gameBoard[x + 1, y + 1] is University)
                {
                    tmpS = (University)_state.gameBoard[x + 1, y + 1];
                    if (tmpS.Id == id) _state.gameBoard[x + 1, y + 1] = new Empty();
                }

                if (_state.gameBoard[x + 1, y - 1] is University)
                {
                    tmpS = (University)_state.gameBoard[x + 1, y - 1];
                    if (tmpS.Id == id) _state.gameBoard[x + 1, y - 1] = new Empty();
                }

                if (_state.gameBoard[x - 1, y - 1] is University)
                {
                    tmpS = (University)_state.gameBoard[x - 1, y - 1];
                    if (tmpS.Id == id) _state.gameBoard[x - 1, y - 1] = new Empty();
                }

                if (_state.gameBoard[x - 1, y + 1] is University)
                {
                    tmpS = (University)_state.gameBoard[x - 1, y + 1];
                    if (tmpS.Id == id) _state.gameBoard[x - 1, y + 1] = new Empty();
                }

                return true;
            }
            else if (tile is School)
            {
                UpdateMaintenanceFees(tile, -School.maintenanceFee);
                _state.money += School.buildCost / 2;
                if (!byCatastrophe) UpdateTransactions(new Transaction(TransactionType.BULLDOZER, School.buildCost / 2, Time, tile.GetType()));
                School tmpS = (School)_state.gameBoard[x, y];
                int id = tmpS.Id;
                _state.gameBoard[x, y] = new Empty();

                if (_state.gameBoard[x + 1, y] is School)
                {
                    tmpS = (School)_state.gameBoard[x + 1, y];
                    if (tmpS.Id == id) _state.gameBoard[x + 1, y] = new Empty();
                }

                if (_state.gameBoard[x - 1, y] is School)
                {
                    tmpS = (School)_state.gameBoard[x - 1, y];
                    if (tmpS.Id == id) _state.gameBoard[x - 1, y] = new Empty();
                }

                if (_state.gameBoard[x, y - 1] is School)
                {
                    tmpS = (School)_state.gameBoard[x, y - 1];
                    if (tmpS.Id == id) _state.gameBoard[x, y - 1] = new Empty();
                }

                if (_state.gameBoard[x, y + 1] is School)
                {
                    tmpS = (School)_state.gameBoard[x, y + 1];
                    if (tmpS.Id == id) _state.gameBoard[x, y + 1] = new Empty();
                }

                if (_state.gameBoard[x + 1, y + 1] is School)
                {
                    tmpS = (School)_state.gameBoard[x + 1, y + 1];
                    if (tmpS.Id == id) _state.gameBoard[x + 1, y + 1] = new Empty();
                }

                if (_state.gameBoard[x + 1, y - 1] is School)
                {
                    tmpS = (School)_state.gameBoard[x + 1, y - 1];
                    if (tmpS.Id == id) _state.gameBoard[x + 1, y - 1] = new Empty();
                }

                if (_state.gameBoard[x - 1, y - 1] is School)
                {
                    tmpS = (School)_state.gameBoard[x - 1, y - 1];
                    if (tmpS.Id == id) _state.gameBoard[x - 1, y - 1] = new Empty();
                }

                if (_state.gameBoard[x - 1, y + 1] is School)
                {
                    tmpS = (School)_state.gameBoard[x - 1, y + 1];
                    if (tmpS.Id == id) _state.gameBoard[x - 1, y + 1] = new Empty();
                }


                return true;
            }
            else if (tile is Forest forest)
            {
                if(!forest.Generated) UpdateMaintenanceFees(tile, -Forest.MaintenanceFee);
                _state.money += Forest.BuildCost / 2;
                _state.gameBoard[x, y] = new Empty();
                if (!byCatastrophe) UpdateTransactions(new Transaction(TransactionType.BULLDOZER, Forest.BuildCost / 2, Time, tile.GetType()));
                return true;
            }
            else
            {
                return false;
            }
        }
        #endregion

        #region Tax Managments
        /// <summary>
        /// Sets the <see cref="ResidentalTax"/>.
        /// </summary>
        /// <param name="tax">new tax precentage</param>
        public void setResidentalTax(int tax)
        {
            if (tax > 0 && tax < 100) _state.ResidentalTax = tax;
            if ((_state.ResidentalTax == 0 && tax == 1) || (_state.ResidentalTax == 1 && tax == 0))  _state.ResidentalTax = tax; 
            if ((_state.ResidentalTax == 100 && tax == 99) || (_state.ResidentalTax == 99 && tax == 100))  _state.ResidentalTax = tax; 
        }
        /// <summary>
        /// Sets the <see cref="ServiceTax"/>.
        /// </summary>
        /// <param name="tax">new tax precentage</param>
        public void setServiceTax(int tax)
        {
            if (tax > 0 && tax < 100) _state.ServiceTax = tax;
            if ((_state.ServiceTax == 0 && tax == 1) || (_state.ServiceTax == 1 && tax == 0)) _state.ServiceTax = tax;
            if ((_state.ServiceTax == 100 && tax == 99) || (_state.ServiceTax == 99 && tax == 100)) _state.ServiceTax = tax;
        }
        /// <summary>
        /// Sets the <see cref="IndustrialTax"/>.
        /// </summary>
        /// <param name="tax">new tax precentage</param>
        public void setIndustrialTax(int tax)
        {
            if (tax > 0 && tax < 100) _state.IndustrialTax = tax;
            if ((_state.IndustrialTax == 0 && tax == 1) || (_state.IndustrialTax == 1 && tax == 0)) _state.IndustrialTax = tax;
            if ((_state.IndustrialTax == 100 && tax == 99) || (_state.IndustrialTax == 99 && tax == 100)) _state.IndustrialTax = tax;
        }
        private void taxCollection()
        {
            var taxIncome = _state.CalculateIncome();
            _state.money += taxIncome.Item1;
            UpdateTransactions(new Transaction(TransactionType.RESIDENTIALTAX, taxIncome.Item1, Time));
            _state.money += taxIncome.Item2;
            UpdateTransactions(new Transaction(TransactionType.SERVICETAX, taxIncome.Item2, Time));
            _state.money += taxIncome.Item3;
            UpdateTransactions(new Transaction(TransactionType.INDUSTRIALTAX, taxIncome.Item3, Time));
        }
        #endregion

        #region Save Load
        /// <summary>
        /// Saves <see cref="MaynotModel"/>.
        /// </summary>
        /// <param name="path">path of file</param>
        public async Task SaveGameAsync(String path)
        {
            if (_dataAccess == null)
                throw new InvalidOperationException("No data access is provided.");

            await _dataAccess.SaveAsync(path, _state);
        }
        /// <summary>
        /// Loads a <see cref="MaynotModel"/>.
        /// </summary>
        /// <remarks>
        /// This method calls the <see cref="OnGameLoaded"/> after the <see cref="_state"/> loads the model.
        /// </remarks>
        /// <param name="path">path of file</param>
        public async Task LoadGameAsync(String path)
		{
			if (_dataAccess == null)
				throw new InvalidOperationException("No data access is provided.");

			InspectedTile = null;
			_state.timer.Elapsed -= Timer_Elapsed;
			_state = await _dataAccess.LoadAsync(path);
			_state.timer.Elapsed += Timer_Elapsed;
			OnGameLoaded();
			Debug.WriteLine("Load game");
		}
        #endregion

        #region Buildings Effects
        /// <summary>
        /// Calculates all <see cref="Tile"/> effects that can change the satisfaction of citizens.
        /// </summary>
        /// <remarks>
        /// This method calls <see cref="industrailZoneEffects"/>, <see cref="policeEffect"/>, <see cref="stadiumEffect"/>, <see cref="forestEffect"/>.
        /// </remarks>
        public void buildingEffects()
        {
            industrailZoneEffects();
            policeEffect();
            stadiumEffect();
            forestEffect();
        }
        private List<Tile> getArea(int size, int x, int y)
        {
            debug++;
            List<Tile> tiels = new List<Tile>();
            //4
            for (int i = 0; i <= 5; ++i)
            {
                for (int j = 1; j <= 5; ++j)
                {
                    if (x + i < _state.size && y + j < _state.size)
                    {
                        tiels.Add(_state.gameBoard[x + i, y + j]);
                    }
                }
            }
            //1
            for (int i = 1; i <= 5; ++i)
            {
                for (int j = 1; j <= 5; ++j)
                {
                    if (x - i >= 0 && y + j < _state.size)
                    {
                        tiels.Add(_state.gameBoard[x - i, y + j]);
                    }
                }
            }
            //2
            for (int i = 1; i <= 5; ++i)
            {
                for (int j = 0; j <= 5; ++j)
                {
                    if (x - i >= 0 && y - j >= 0)
                    {
                        tiels.Add(_state.gameBoard[x - i, y - j]);
                    }
                }
            }
            //3
            for (int i = 0; i <= 5; ++i)
            {
                for (int j = 0; j <= 5; ++j)
                {
                    if (x + i < _state.size && y - j >= 0)
                    {
                        tiels.Add(_state.gameBoard[x + i, y - j]);
                    }             
                }
            }

            return tiels;
        }
        private void industrailZoneEffects()
        {
            List<Tile> tiels = new List<Tile>();
            for (int i = 0; i < _state.industrialZones.Count; ++i)
            {
				int x = _state.industrialZones[i].Item1;
				int y = _state.industrialZones[i].Item2;

				if (_state.IsDistance(x, y, GetStartRoadX(), GetStartRoadY()) == 0)
                    return;

                List<ResidentialZone> rez = new List<ResidentialZone>();
                tiels = getArea(5, x, y);
                for (int j = 0; j < tiels.Count; ++j)
                {
                    if (tiels[j] is MaynotPersistence.ResidentialZone)
                    {
                        rez.Add((ResidentialZone)tiels[j]);
                    }
                }
				for (int j = 0; j < rez.Count; ++j)
				{
					for (int k = 0; k < rez[j].GetPeoples(_state.citizens).Count; ++k)
					{
                        if (rez[j].GetPeoples(_state.citizens)[k].Satisfaction >= 5)
                        {
                            rez[j].GetPeoples(_state.citizens)[k].Satisfaction -= 5;
                        }
                    }
				}
            }
        }
        private void policeEffect()
        {
            List<Tile> tiels = new List<Tile>();
            for (int i = 0; i < _state.size; ++i)
            {
                for (int k = 0; k < _state.size; ++k)
                {
                    if (_state.gameBoard[i, k] is PoliceStation)
                    {
                        List<ResidentialZone> rez = new List<ResidentialZone>();
                        int x = i;
                        int y = k;

						if (_state.IsDistance(x, y, GetStartRoadX(), GetStartRoadY()) == 0)
							return;

						tiels = getArea(5, x, y);
                        for (int j = 0; j < tiels.Count; ++j)
                        {
                            if (tiels[j] is MaynotPersistence.ResidentialZone)
                            {
                                rez.Add((ResidentialZone)tiels[j]);
                            }
                        }
                        for (int j = 0; j < rez.Count; ++j)
                        {
                            for (int l = 0; l < rez[j].GetPeoples(_state.citizens).Count; ++l)
                            {
                                if (rez[j].GetPeoples(_state.citizens)[l].Satisfaction <= 95)
                                {
                                    rez[j].GetPeoples(_state.citizens)[l].Satisfaction += 5;
                                }
                                else
                                {
                                    rez[j].GetPeoples(_state.citizens)[l].Satisfaction = 100;
                                }
                            }
                        }
                    }
                }             
            }
        }
        private void stadiumEffect()
        {
            List<Tile> tiels = new List<Tile>();
            for (int i = 0; i < _state.size; ++i)
            {
                for (int k = 0; k < _state.size; ++k)
                {
                    if (_state.gameBoard[i, k] is Stadium)
                    {
                        List<ResidentialZone> rez = new List<ResidentialZone>();
                        int x = i;
                        int y = k;

						if (_state.IsDistance(x, y, GetStartRoadX(), GetStartRoadY()) == 0)
							return;

						tiels = getArea(5, x, y);
                        for (int j = 0; j < tiels.Count; ++j)
                        {
                            if (tiels[j] is MaynotPersistence.ResidentialZone)
                            {
                                rez.Add((ResidentialZone)tiels[j]);
                            }
                        }
                        for (int j = 0; j < rez.Count; ++j)
                        {
                            for (int l = 0; l < rez[j].GetPeoples(_state.citizens).Count; ++l)
                            {
                                if (rez[j].GetPeoples(_state.citizens)[l].Satisfaction <= 95)
                                {
                                    rez[j].GetPeoples(_state.citizens)[l].Satisfaction += 5;
                                }
                                else
                                {
                                    rez[j].GetPeoples(_state.citizens)[l].Satisfaction = 100;
                                }
                            }
                        }
                    }
                }
            }
        }
        private void forestEffect()
        {
            List<Tile> tiels = new List<Tile>();
            for (int i = 0; i < _state.size; ++i)
            {
                for (int k = 0; k < _state.size; ++k)
                {
                    if (_state.gameBoard[i, k] is Forest forest)
                    {
                        List<ResidentialZone> rez = new List<ResidentialZone>();
                        int x = i;
                        int y = k;
                        int value = Math.Min((int)Math.Floor(0.5 * forest.Age), 5);

						tiels = getArea(5, x, y);
                        for (int j = 0; j < tiels.Count; ++j)
                        {
                            if (tiels[j] is MaynotPersistence.ResidentialZone)
                            {
                                rez.Add((ResidentialZone)tiels[j]);
                            }
                        }
						for (int j = 0; j < rez.Count; ++j)
						{
							for (int l = 0; l < rez[j].GetPeoples(_state.citizens).Count; ++l)
							{
                                if (rez[j].GetPeoples(_state.citizens)[l].Satisfaction <= 95)
                                {
                                    rez[j].GetPeoples(_state.citizens)[l].Satisfaction += value;
                                }
                                else
                                {
                                    rez[j].GetPeoples(_state.citizens)[l].Satisfaction = 100;
                                }
                            }
						}						
                    }
                }
            }
        }
		#endregion

		#region Buisness Logics
		private void catastrophe()
		{
			Random happens = new Random();
			int chance = happens.Next(0, 100);
			if (chance == 0)
			{
				int x;
				int y;
				x = happens.Next(0, 30);
				y = happens.Next(0, 30);
				_state.gameBoard[x, y] = new Empty();
				List<Tile> tiles = new List<Tile>();
				//tiles = getArea(5,x,y);

				//_state.gameBoard[x + 1, y] = new PoliceStation();
				//4
				for (int i = 0; i <= 5; ++i)
				{
					for (int j = 1; j <= 5; ++j)
					{
						if (x + i < _state.size && y + j < _state.size)
						{
							DestroyTile(x + i, y + j, true);
							//_state.gameBoard[] = new Empty();
						}
					}
				}
				//1
				for (int i = 1; i <= 5; ++i)
				{
					for (int j = 1; j <= 5; ++j)
					{
						if (x - i >= 0 && y + j < _state.size)
						{
							DestroyTile(x - i, y + j, true);
							//_state.gameBoard[x - i, y + j] = new Empty();
						}
					}
				}
				//2
				for (int i = 1; i <= 5; ++i)
				{
					for (int j = 0; j <= 5; ++j)
					{
						if (x - i >= 0 && y - j >= 0)
						{
							DestroyTile(x - i, y - j, true);
							//_state.gameBoard[x - i, y - j] = new Empty();
						}
					}
				}
				//3
				for (int i = 0; i <= 5; ++i)
				{
					for (int j = 0; j <= 5; ++j)
					{
						if (x + i < _state.size && y - j >= 0)
						{
							DestroyTile(x + i, y - j, true);
							//_state.gameBoard[x + i, y - j] = new Empty();
						}
					}
				}


				// _state.gameBoard[x + 1, y] = new Stadium();



				catastropheHappened?.Invoke(this, new MaynotEventArg(x, y));
			}
		}
		private void aging()
		{
			for (int i = 0; i < _state.citizens.Count; ++i)
			{
				_state.citizens[i].Age += 1;
                if (_state.citizens[i].Age == 65)
                {
					_state.citizens[i].Education = Level.RETIRED;
					_state.citizens[i].WorkPlace = (-1, -1);
				}                    
			}
			OnPopulationUpdated();
		}
		private void MovingIn()
		{
			Random r = new();
			int newCitizens = r.Next(0, 25);
			bool anyUpdates = false;
			for (int i = 0; i < newCitizens; i++)
			{
				anyUpdates = _state.GenerateNewPerson(60);

				if (Citizens.Count >= 50 && _state.guaranteedPopulation)
					_state.guaranteedPopulation = false;
			}
			if (anyUpdates)
				OnPopulationUpdated();
		}
        private void MovingOut()
        {
			for (int i = 0; i < Citizens.Count; i++)
			{
				if (!_state.guaranteedPopulation && _state.CheckMovingOut(Citizens[i]))
				{
					_state.RemovePerson(Citizens[i]);
				}
			}
			OnPopulationUpdated();
		}
        /// <summary>
        /// Sets taxes payed for each citizen 
        /// and calculates the total summary of taxes.
        /// </summary>
        /// <remarks>
        /// This method changes the value of the <see cref="Transactions"/> property.
        /// </remarks>
        public void managingPension()
		{
			int currAge = 0;
			for (int i = 0; i < Citizens.Count; ++i)
			{
				currAge = Citizens[i].Age;
				if (20 > 60 - currAge && 60 - currAge > 0)
				{
					Citizens[i].Taxes[60 - currAge] = _state.ResidentalTax;

				}
			}
            float pensionSum = 0;
			for (int i = 0; i < Citizens.Count; ++i)
			{
				if (Citizens[i].Age > 60)
				{
                    pensionSum += Citizens[i].Taxes.Average();
				}
			}
            _state.money -= pensionSum;
            UpdateTransactions(new Transaction(TransactionType.PENSION, pensionSum, Time));
        }
        private void CalculateAverageSatisfaction()
		{
			int sum = 0;
			for (int i = 0; i < _state.citizens.Count; ++i)
			{
				sum += _state.citizens[i].Satisfaction;
			}
			if (_state.citizens.Count != 0)
			{
				_state.averageSatisfaction = sum / _state.citizens.Count;
			}
			else
			{
				_state.averageSatisfaction = 0;
			}

		}
		private void growForests()
        {
            for (int i = 0; i < _state.size; ++i)
            {
                for (int k = 0; k < _state.size; ++k)
                {
                    if (_state.gameBoard[i, k] is Forest)
                    {
                        Forest? f = _state.gameBoard[i, k] as Forest;
                        if (f != null && (f.PlantingDate - Time).TotalDays % 365 == 0 && f.Age < 10)
                        {
                            f.Age++;
                            if(f.Age == 10 && !f.Generated)
                            {
                                UpdateMaintenanceFees(f, -Forest.MaintenanceFee);
                            }
                        }
                    }
                }
            }
        }
        private void EndLifes()
        {
			for (int i = 0; i < Citizens.Count; i++)
			{
				if (_state.CheckMortality(Citizens[i]))
				{
                    _state.RemovePerson(Citizens[i]);
					_state.GenerateNewPerson(18);
					OnPopulationUpdated();
				}
			}
        }
        private void SetEducationLevel()
        {
            int numberOfSchools = _state.CountSchools();
            int numberOfUniversities = _state.CountUniversities();
            int numberOfIntermediates = _state.CountIntermediates();
            int numberOfSuperlatives = _state.CountSuperlatives();
            int maxIntermediates = Math.Min((int)(Citizens.Count * 0.05 * numberOfSchools), (int)(Citizens.Count * 0.4));
			int maxSuperlatives = Math.Min((int)(Citizens.Count * 0.05 * numberOfUniversities), (int)(Citizens.Count * 0.3));

            for (int i = 0; i < numberOfSchools; i++)
                _state.SetIntermediatesLevel(numberOfIntermediates, maxIntermediates);

            for (int i = 0; i < numberOfUniversities; i++)
                _state.SetRuperlativesLevel(numberOfSuperlatives, maxSuperlatives);
		}

        private void negativeMoney()
        {
            if (_state.money < 0)
            {
                int dissatisfaction;
                _state.negative++;
                dissatisfaction = (int)_state.money / 1000;
                dissatisfaction = dissatisfaction * -1;
                dissatisfaction = dissatisfaction * _state.negative / 3;
                for (int i = 0; i < _state.citizens.Count; ++i)
                {
                    if (_state.citizens[i].Satisfaction >= dissatisfaction)
                    {
                        _state.citizens[i].Satisfaction -= dissatisfaction;
                    }
                    else
                    {
                        _state.citizens[i].Satisfaction = 0;
                    }
                }
            }
            else
            {
                _state.negative = 0;
            }
        }

        private void highTaxes()
        {
            if (_state.ResidentalTax > 75)
            {
                for (int i = 0; i < _state.citizens.Count; ++i)
                {
                    _state.citizens[i].Satisfaction -= 5;
                }
            }
            if (_state.IndustrialTax > 90)
            {
                for (int i = 0; i < _state.citizens.Count; ++i)
                {
                    _state.citizens[i].Satisfaction -= 5;
                }
            }
            if (_state.ServiceTax > 90)
            {
                for (int i = 0; i < _state.citizens.Count; ++i)
                {
                    _state.citizens[i].Satisfaction -= 5;
                }
            }
        }

        private void lowTaxes()
        {
            if (_state.ResidentalTax < 35)
            {
                for (int i = 0; i < _state.citizens.Count; ++i)
                {
                    _state.citizens[i].Satisfaction += 3;
                }
            }
            if (_state.IndustrialTax < 50)
            {
                for (int i = 0; i < _state.citizens.Count; ++i)
                {
                    _state.citizens[i].Satisfaction += 3;
                }
            }
            if (_state.ServiceTax < 50)
            {
                for (int i = 0; i < _state.citizens.Count; ++i)
                {
                    _state.citizens[i].Satisfaction += 3;
                }
            }
        }

        private void lose()
        {
            if (_state.averageSatisfaction < 10 || _state.citizens.Count == 0)
            {
                stopTime();
                GameOver?.Invoke(this, new EventArgs());
            }
        }
        private void PayExpenses()
        {
            foreach (Type tileType in _state.MaintenanceFees.Keys)
            {
                _state.money -= _state.MaintenanceFees[tileType];
                if (_state.MaintenanceFees[tileType] != 0)
                {
                    UpdateTransactions(new Transaction(TransactionType.MAINTENANCE, _state.MaintenanceFees[tileType], Time, tileType));
                }
            }
        }
        private void UpdateMaintenanceFees(Tile tileType, float fee)
        {
            if (_state.MaintenanceFees.ContainsKey(tileType.GetType()))
            {
                if (fee < 0 && _state.MaintenanceFees[tileType.GetType()] + fee < 0)
                {
                    throw new Exception("Can't have negative fee for a tile type");
                }
                _state.MaintenanceFees[tileType.GetType()] += fee;
            }
            else
            {
                _state.MaintenanceFees[tileType.GetType()] = fee;
            }
        }
        #endregion

        #region Private event methods
        /// <summary>
        /// Invokes DayElapsed event
        /// </summary>
        private void OnTimeElapsed()
		{
			DayElapsed?.Invoke(this, new TimeElapsedEventArgs(_state.time));
		}
        /// <summary>
        /// Invokes UpdatePopulation event
        /// </summary>
		private void OnPopulationUpdated()
		{
			UpdatePopulation?.Invoke(this, new EventArgs());
		}
        /// <summary>
        /// Invokes GameSpeedChanged event
        /// </summary>
		private void OnGameSpeedChanged()
		{
			GameSpeedChanged?.Invoke(this, new EventArgs());
		}
        
        private void OnGameLoaded()
		{
			OnGameSpeedChanged();
			OnInspectedTileChanged();
			OnTimeElapsed();
			OnPopulationUpdated();
			GameLoaded?.Invoke(this, new MaynotEventArg());
		}
        /// <summary>
        /// Invokes InspectedTileChanged event
        /// </summary>
		private void OnInspectedTileChanged()
		{
			InspectedTileChanged?.Invoke(this, new InspectedTileChangedEventArgs(InspectedTile));
		}
        #endregion

        /// <summary>
        /// Update list of saved transactions with
        /// <paramref name="newTransaction"/>
        /// </summary>
        private void UpdateTransactions(Transaction newTransaction)
        {
            if (newTransaction.TransactionAmount == 0) return;
            while (Transactions.Count >= 200) Transactions.RemoveAt(0);
            Transactions.Add(newTransaction);
            TransactionHappenned?.Invoke(this, new TransactionEventArgs(newTransaction));
        }
    }
}
