﻿using System;
using System.Collections.Generic;
using System.Diagnostics;

namespace MaynotPersistence
{
	public class MaynotGameState
	{
		#region Properties
		public System.Timers.Timer timer;
		public DateTime time;

		public Tile[,] gameBoard;
		public bool[,] visited;

		public int size;
		public int startX;
		public int startY;
		public int gameSpeed;
		public int prevGameSpeed;
		public int yearTracker;
		public int weakTracker;
		public bool guaranteedPopulation = true;

		public float income;
		public float expense;
        public Dictionary<Type, float> MaintenanceFees { get; set; }
        public float money;
		public int negative;

		public List<Person> citizens;
		public List<(int, int)> residentalZones;
		public List<(int, int)> serviceZones;
		public List<(int, int)> industrialZones;

		public int ResidentalTax;
		public int IndustrialTax;
		public int ServiceTax;
		public double averageSatisfaction;

		private const float _TaxMultElemLvl = 1.0F;
		private const float _TaxMultInterlLvl = 1.6F;
		private const float _TaxMulSuperLvl = 2.3F;
		#endregion

		#region Constructors
		public MaynotGameState()
		{
			timer = new System.Timers.Timer(500);
			time = new DateTime(1800, 1, 1);

			gameBoard = new Tile[0, 0];
			visited = new bool[0, 0];

			citizens = new List<Person>();
			residentalZones = new List<(int, int)>();
			serviceZones = new List<(int, int)>();
			industrialZones = new List<(int, int)>();
		}
		public MaynotGameState(int boardSize)
		{
			timer = new System.Timers.Timer(500);
			time = new DateTime(1800, 1, 1);

			gameBoard = new Tile[boardSize, boardSize];
			visited = new bool[boardSize, boardSize];

			money = 100000;
			MaintenanceFees = new Dictionary<Type, float>();
			gameSpeed = 0;
			yearTracker = 1;
			weakTracker = 0;
			guaranteedPopulation = true;
			ResidentalTax = 69;
			ServiceTax = 4;
			IndustrialTax = 95;
			size = boardSize;
			startX = size / 2 - 1;
			startY = 0;

			for (int i = 0; i < size; i++)
			{
				for (int j = 0; j < size; j++)
				{
					gameBoard[i, j] = new Empty();
				}
			}

			gameBoard[startX, startY] = new Road();

			citizens = new List<Person>();
			residentalZones = new List<(int, int)>();
			serviceZones = new List<(int, int)>();
			industrialZones = new List<(int, int)>();

			SetVisitedEmpty();
		}
		#endregion

		#region Path network helper functions
		/// <summary>
		/// Beállítja a út-kereséshez a logikai táblát.
		/// </summary>
		public void SetVisitedEmpty()
		{
			for (int i = 0; i < size; i++)
				for (int j = 0; j < size; j++)
					visited[i, j] = false;
		}
		/// <summary>
		/// Megnézi, hogy a paraméterül kapott pozíción épület van-e.
		/// <paramref name="i"/> első koordináta.
		/// <paramref name="j"/> második koordináta.
		/// <returns>Ha épület akkor igaz, egyébként hamis.</returns>
		/// </summary>
		public bool IsBuilding(int i, int j)
		{
			if (!IsValid(i, j))
				return false;

			return gameBoard[i, j] is Zone || gameBoard[i, j] is Facility;
		}
		/// <summary>
		/// Megnézi, hogy a paraméterül kapott pozíció a pályán belül van-e.
		/// <paramref name="i"/> első koordináta.
		/// <paramref name="j"/> második koordináta.
		/// <returns>Ha belül van akkor igaz, egyébként hamis.</returns>
		/// </summary>
		public bool IsValid(int i, int j)
		{
			return i >= 0 && i < size && j >= 0 && j < size;
		}
		/// <summary>
		/// Megnézi, hogy a paraméterül kapott pozíción út van-e.
		/// <paramref name="i"/> első koordináta.
		/// <paramref name="j"/> második koordináta.
		/// <returns>Ha út akkor igaz, egyébként hamis.</returns>
		/// </summary>
		public bool IsRoad(int i, int j)
		{
			if (IsValid(i, j))
				return gameBoard[i, j] is Road;

			return false;
		}
		private int Distance(int i, int j, int x, int y)
		{
			if (IsValid(i, j) && IsValid(x, y) && !visited[i, j])
			{
				visited[i, j] = true;

				if (IsValid(i - 1, j) && i - 1 == x && j == y)
				{
					return 1;
				}
				else
				{
					if (IsRoad(i - 1, j))
					{
						int up = Distance(i - 1, j, x, y);
						if (up > 0)
							return up + 1;
					}
				}

				if (IsValid(i, j - 1) && i == x && j - 1 == y)
				{
					return 1;
				}
				else
				{
					if (IsRoad(i, j - 1))
					{
						int left = Distance(i, j - 1, x, y);
						if (left > 0)
							return left + 1;
					}
				}

				if (IsValid(i + 1, j) && i + 1 == x && j == y)
				{
					return 1;
				}
				else
				{
					if (IsRoad(i + 1, j))
					{
						int down = Distance(i + 1, j, x, y);
						if (down > 0)
							return down + 1;
					}
				}

				if (IsValid(i, j + 1) && i == x && j + 1 == y)
				{
					return 1;
				}
				else
				{
					if (IsRoad(i, j + 1))
					{
						int right = Distance(i, j + 1, x, y);
						if (right > 0)
							return right + 1;
					}
				}
			}

			return 0;
		}
		public int IsDistance(int i, int j, int x, int y)
		{
			SetVisitedEmpty();
			return Distance(i, j, x, y);
		}
		private List<Tuple<int, int>> ReachableBuildings(int i, int j, int x, int y)
		{
			List<Tuple<int, int>> reachableBuildingsList = new();
			if (IsValid(i, j) && !visited[i, j])
			{
				visited[i, j] = true;

				if (IsBuilding(i - 1, j))
				{
					reachableBuildingsList.Add(new Tuple<int, int>(i - 1, j));
				}
				else if (IsRoad(i - 1, j))
				{
					if (i - 1 != x || j != y)
						reachableBuildingsList.AddRange(ReachableBuildings(i - 1, j, x, y));
				}

				if (IsBuilding(i, j - 1))
				{
					reachableBuildingsList.Add(new Tuple<int, int>(i, j - 1));
				}
				else if (IsRoad(i, j - 1))
				{
					if (i != x || j - 1 != y)
						reachableBuildingsList.AddRange(ReachableBuildings(i, j - 1, x, y));
				}

				if (IsBuilding(i + 1, j))
				{
					reachableBuildingsList.Add(new Tuple<int, int>(i + 1, j));
				}
				else if (IsRoad(i + 1, j))
				{
					if (i + 1 != x || j != y)
						reachableBuildingsList.AddRange(ReachableBuildings(i + 1, j, x, y));
				}

				if (IsBuilding(i, j + 1))
				{
					reachableBuildingsList.Add(new Tuple<int, int>(i, j + 1));
				}
				else if (IsRoad(i, j + 1))
				{
					if (i != x || j + 1 != y)
						reachableBuildingsList.AddRange(ReachableBuildings(i, j + 1, x, y));
				}
			}

			return reachableBuildingsList;
		}
		private List<Tuple<int, int>> GetReachableBuildings(int i, int j)
		{
			SetVisitedEmpty();
			return ReachableBuildings(i, j, size, size);
		}
		private List<Tuple<int, int>> GetReachableBuildingsFromStart(int x, int y)
		{
			SetVisitedEmpty();
			return ReachableBuildings(14, 0, x, y);
		}
		private List<Tuple<int, int>> GetReachableBuildingsIntercept(int i, int j)
		{
			List<Tuple<int, int>> ReachableBuildingsFromCell = GetReachableBuildings(i, j);
			List<Tuple<int, int>> ReachableBuildingsFromStart = GetReachableBuildingsFromStart(i, j);
			ReachableBuildingsFromCell.RemoveAll(t => ReachableBuildingsFromStart.Contains(t));
			return ReachableBuildingsFromCell;
		}
		#endregion

		#region Tax helper functions		
		public float CalculateResidentalTax()
		{
			List<(int x, int y)> vResidentalZones = residentalZones;
			Tile[,] vGameBoard = gameBoard;
			float residentalTax = 0;

			foreach ((int x, int y) in vResidentalZones)
			{
				if (vGameBoard[x, y] is ResidentialZone z)
				{
					foreach (Person p in z.GetPeoples(citizens))
					{
						residentalTax += ResidentalTax;
						//TODO: befizetett adó tárolása személyeknél a nyugdíj miatt
					}
				}
			}

			return residentalTax;
		}
		public float CalculateServiceTax()
		{
			List<(int x, int y)> vServiceZones = serviceZones;
			Tile[,] vGameBoard = gameBoard;
			float serviceTax = 0;

			foreach ((int x, int y) in vServiceZones)
			{
				if (vGameBoard[x, y] is ServiceZone z)
				{
					foreach (Person p in z.GetPeoples(citizens))
					{
						if (p.Education == Level.ELEMENTARY)
						{
							serviceTax += ServiceTax * _TaxMultElemLvl;
							//TODO: befizetett adó tárolása személyeknél a nyugdíj miatt
						}

						if (p.Education == Level.INTERMEDIATE)
						{
							serviceTax += ServiceTax * _TaxMultInterlLvl;
							//TODO: befizetett adó tárolása személyeknél a nyugdíj miatt
						}

						if (p.Education == Level.SUPERLATIVE)
						{
							serviceTax += ServiceTax * _TaxMulSuperLvl;
							//TODO: befizetett adó tárolása személyeknél a nyugdíj miatt
						}
					}
				}
			}

			return serviceTax;
		}
		public float CalculateIndustrialTax()
		{
			List<(int x, int y)> vIndustrialZones = industrialZones;
			Tile[,] vGameBoard = gameBoard;
			float industrialTax = 0;

			foreach ((int x, int y) in vIndustrialZones)
			{
				if (vGameBoard[x, y] is IndustrialZone z)
				{
					foreach (Person p in z.GetPeoples(citizens))
					{
						if (p.Education == Level.ELEMENTARY)
						{
							industrialTax += IndustrialTax * _TaxMultElemLvl;
							//TODO: befizetett adó tárolása személyeknél a nyugdíj miatt
						}

						if (p.Education == Level.INTERMEDIATE)
						{
							industrialTax += IndustrialTax * _TaxMultInterlLvl;
							//TODO: befizetett adó tárolása személyeknél a nyugdíj miatt
						}

						if (p.Education == Level.SUPERLATIVE)
						{
							industrialTax += IndustrialTax * _TaxMulSuperLvl;
							//TODO: befizetett adó tárolása személyeknél a nyugdíj miatt
						}
					}
				}
			}

			return industrialTax;
		}
		public Tuple<float,float,float> CalculateIncome()
		{
			return new Tuple<float,float,float>(CalculateResidentalTax(), CalculateServiceTax(), CalculateIndustrialTax());
		}
		#endregion

		#region Move-in helper functions
		public double RatioIndustrialZonesFullness()
		{
			int counterOccupied = 0;
			int counterAll = 0;
			foreach ((int x, int y) in industrialZones)
			{
				if (gameBoard[x, y] is Zone z && IsDistance(x, y, startX, startY) != 0)
				{
					counterOccupied += z.GetPeoples(citizens).Count;
					counterAll += z.Capacity;
				}
			}

			if (counterAll != 0)
			{
				return (double)counterOccupied / counterAll;
			}

			return 1;
		}
		public double RatioServiceZonesFullness()
		{
			int counterOccupied = 0;
			int counterAll = 0;
			foreach ((int x, int y) in serviceZones)
			{
				if (gameBoard[x, y] is Zone z && IsDistance(x, y, startX, startY) != 0)
				{
					counterOccupied += z.GetPeoples(citizens).Count;
					counterAll += z.Capacity;
				}
			}

			if (counterAll != 0)
			{
				return (double)counterOccupied / counterAll;
			}

			return 1;
		}
		public (int x, int y) GetFreeResidentalZone()
		{
			Random r = new();
			List<(int x, int y)> freeResidentialZones = new();
			foreach ((int x, int y) in residentalZones)
			{
				if (gameBoard[x, y] is Zone z && z.Capacity > z.GetPeoples(citizens).Count && IsDistance(x, y, startX, startY) > 0)
				{
					freeResidentialZones.Add((x, y));
				}
			}

			if (freeResidentialZones.Count > 0)
			{
				int randomIndex = r.Next(0, freeResidentialZones.Count);
				return freeResidentialZones[randomIndex];
			}

			return (-1, -1);
		}
		public (int x, int y) GetFreeServiceZone(int i, int j)
		{
			Random r = new();
			List<(int x, int y, int d)> freeServiceZones = new();
			foreach ((int x, int y) in serviceZones)
			{
				if (gameBoard[x, y] is Zone z && z.Capacity > z.GetPeoples(citizens).Count && IsDistance(x, y, i, j) > 0
					&& IsDistance(x, y, startX, startY) > 0)
				{
					freeServiceZones.Add((x, y, IsDistance(x, y, i, j)));
				}
			}

			if (freeServiceZones.Count > 0)
			{
				freeServiceZones.Sort((a, b) => a.d.CompareTo(b.d));
				return (freeServiceZones[0].x, freeServiceZones[0].y);
			}

			return (-1, -1);
		}
		public (int x, int y) GetFreeIndustrialZone(int i, int j)
		{
			Random r = new();
			List<(int x, int y, int d)> freeIndustrialZones = new();
			foreach ((int x, int y) in industrialZones)
			{
				if (gameBoard[x, y] is Zone z && z.Capacity > z.GetPeoples(citizens).Count && IsDistance(x, y, i, j) > 0
					&& IsDistance(x, y, startX, startY) > 0)
				{
					freeIndustrialZones.Add((x, y, IsDistance(x, y, i, j)));
				}
			}

			if (freeIndustrialZones.Count > 0)
			{
				freeIndustrialZones.Sort((a, b) => a.d.CompareTo(b.d));
				return (freeIndustrialZones[0].x, freeIndustrialZones[0].y);
			}

			return (-1, -1);
		}
		public (int x, int y) GetFreeWorkZone(int i, int j)
		{
			(int x, int y) workPlace;
			if (RatioServiceZonesFullness() <= RatioIndustrialZonesFullness())
			{
				workPlace = GetFreeServiceZone(i, j);
			}
			else
			{
				workPlace = GetFreeIndustrialZone(i, j);
			}

			return workPlace;
		}
		public int FindClosestIndustrialZone(int i, int j)
		{
			int circleSize = 4;
			List<(int, int)> iZones = new();

			for (int k = i - circleSize; k <= i + circleSize; k++)
			{
				for (int l = j - circleSize; l <= j + circleSize; l++)
				{
					if (IsValid(k, l) && gameBoard[k, l] is IndustrialZone)
					{
						iZones.Add((k, l));
					}
				}
			}

			int minDistance = 0;
			if (iZones.Count > 0)
			{
				minDistance = (int)Math.Sqrt((iZones[0].Item1 - i) * (iZones[0].Item1 - i) + (iZones[0].Item2 - j) * (iZones[0].Item2 - j));
				foreach ((int, int) Coord in iZones)
				{
					int distance = (int)Math.Sqrt((Coord.Item1 - i) * (Coord.Item1 - i) + (Coord.Item2 - j) * (Coord.Item2 - j));
					if (distance < minDistance)
					{
						minDistance = distance;
					}
				}
			}

			if (minDistance == 0)
				return minDistance;

			return 6 - minDistance;
		}
		public int CountForests(int i, int j)
		{
			int circleSize = 3;
			int counter = 0;

			for (int k = i - circleSize; k <= i + circleSize; k++)
			{
				for (int l = j - circleSize; l <= j + circleSize; l++)
				{
					if (IsValid(k, l) && gameBoard[k, l] is Forest)
					{
						counter++;
					}
				}
			}

			return counter;
		}
		public int CalculateZoneAttractiveness(Person p)
		{
			int attractiveness = 100;
			if (CalculateAverageSatisfaction() == 0)
				attractiveness -= 20;
			else
				attractiveness -= (int)CalculateAverageSatisfaction();
			attractiveness -= 2 * CountForests(p.Residency.Item1, p.Residency.Item2);
			attractiveness += IsDistance(p.Residency.Item1, p.Residency.Item2, p.WorkPlace.Item1, p.WorkPlace.Item2);
			attractiveness += (int)Math.Pow(3, FindClosestIndustrialZone(p.Residency.Item1, p.Residency.Item2));
			return attractiveness;
		}
		#endregion

		#region Peoples helper functions
		public bool GenerateNewPerson(int maxAge)
		{
			Random r = new Random();
			Person? p = null;
			int age = r.Next(18, maxAge);
			int satisfaction = r.Next(30, 100);
			int entryValue = r.Next(satisfaction, 100);
			(int, int) t = GetFreeResidentalZone();
			(int, int) w = GetFreeWorkZone(t.Item1, t.Item2);

			if (t.Item1 != -1 && t.Item2 != -1 && w.Item1 != -1 && w.Item2 != -1)
			{
				p = new Person(satisfaction, age, t, w, Level.ELEMENTARY);
			}

			if (p != null && (guaranteedPopulation || CalculateZoneAttractiveness(p) < entryValue))
			{
				if (gameBoard[t.Item1, t.Item2] is Zone && gameBoard[w.Item1, w.Item2] is Zone)
				{
					citizens.Add(p);
					return true;
				}
			}

			return false;
		}
		public float CalculateAverageSatisfaction()
		{
			List<Person> vCitizens = citizens;
			float averageSatisfaction = 0;

			for (int i = 0; i < citizens.Count; ++i)
			{
				averageSatisfaction += citizens[i].Satisfaction;
			}

			/*foreach (Person p in vCitizens)
			{
				averageSatisfaction += p.Satisfaction;
			}*/

			if (citizens.Count == 0)
				return averageSatisfaction;

			return averageSatisfaction / citizens.Count;
		}
		public bool CheckMortality(Person p)
		{
			Random r = new Random();
			int value = (int)Math.Pow(1.14062, p.Age - 65);
			int chance = r.Next(1, 101);

			return value > chance;
		}
		public bool CheckMovingOut(Person p)
		{
			Random r = new Random();
			int value = (int)Math.Pow(1.166, 30 - p.Satisfaction);
			int chance = r.Next(1, 101);

			return value > chance;
		}
		public bool RemovePersonFromCitizens(Person p)
		{
			if (p.Age < 65 && citizens.Remove(p))
				return true;

			return false;
		}
		public bool RemovePerson(Person p)
		{
			if (!RemovePersonFromCitizens(p))
				return false;

			return true;
		}
		#endregion

		#region Destroy Items helper function
		public bool CanDestroyRoad(int i, int j)
		{
			if (i == startX && j == startY)
				return false;

			if (GetReachableBuildingsIntercept(i, j).Count == 0 || IsDistance(i, j, startX, startY) == 0)
				return true;

			return false;
		}
		#endregion

		#region Education helper functions
		public int CountSchools()
		{
			int counter = 0;
			for (int i = 0; i < size; i++)
				for (int j = 0; j < size; j++)
					if (gameBoard[i, j] is School)
						counter++;

			return counter / 2;
		}
		public int CountUniversities()
		{
			int counter = 0;
			for (int i = 0; i < size; i++)
				for (int j = 0; j < size; j++)
					if (gameBoard[i, j] is University)
						counter++;

			return counter / 4;
		}
		public int CountIntermediates()
		{
			int counter = 0;
			foreach (Person p in citizens)
				if (p.Education == Level.INTERMEDIATE)
					counter++;

			return counter;
		}
		public int CountSuperlatives()
		{
			int counter = 0;
			foreach (Person p in citizens)
				if (p.Education == Level.SUPERLATIVE)
					counter++;

			return counter;
		}
		public void SetIntermediatesLevel(int numberOfIntermediates, int maxIntermediates)
		{
			Random r = new Random();
			int HowMany = r.Next(100, 250);
			int Index = r.Next(0, citizens.Count);
			int Counter = 0;

			while (Counter != HowMany && Counter != maxIntermediates - numberOfIntermediates)
			{
				if (citizens[Index].Education == Level.ELEMENTARY)
				{
					citizens[Index].Education = Level.INTERMEDIATE;
					Counter++;
				}
				Index = r.Next(0, citizens.Count);
			}
		}
		public void SetRuperlativesLevel(int numberOfSuperlatives, int maxSuperlatives)
		{
			Random r = new Random();
			int HowMany = r.Next(50, 200);
			int Index = r.Next(0, citizens.Count);
			int Counter = 0;

			while(Counter != HowMany && Counter != maxSuperlatives - numberOfSuperlatives)
			{
				if (citizens[Index].Education == Level.INTERMEDIATE)
				{
					citizens[Index].Education = Level.SUPERLATIVE;
					Counter++;
				}
				Index = r.Next(0, citizens.Count);
			}
		}

		#endregion
	}
}
